/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.notification.tool;

import org.apache.velocity.VelocityContext;
import org.apache.ws.notification.base.v2004_06.impl.WsnNamespaceVersionHolderImpl;
import org.apache.ws.notification.base.v2004_06.porttype.NotificationProducerPortType;
import org.apache.ws.notification.tool.v2004_06.NotificationConsumerPortType2JavaInfo;
import org.apache.ws.notification.tool.v2004_06.NotificationProducerPortType2JavaInfo;
import org.apache.ws.notification.tool.v2004_06.SubscriptionManagerPortType2JavaInfo;
import org.apache.ws.resource.ResourceDefinition;
import org.apache.ws.resource.tool.GenerationUtils;
import org.apache.ws.resource.tool.Wsdl2Java;
import org.apache.ws.resource.tool.velocity.ServiceProperties;
import javax.xml.namespace.QName;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Generates Java Web service types and skeletons from a WSRF/WSN WSDL.
 *
 * @author Ian Springer
 */
public class WsnWsdl2Java
   extends Wsdl2Java
{
   private static final String VCTY_CTX_PROP_CHANGE                  = "propChange";
   private static final String VCTY_CTX_RESOURCE_TERMINATION         = "resourceTermination";
   private static final String VCTY_CTX_NOTIF_PRODUCER               = "notificationProducer";
   private static final String VCTY_CTX_TOPICS_XML_RESOURCE_PATHS    = "propValueChangeTopicsXmlResourcePaths";
   private static final String WEBAPP_NAME                           = "pubscribe";
   private static final String FILENAME_PROP_VALUE_CHANGE_TOPICS_XML = "ResourcePropertyValueChangeTopics.xml";

   /**
    * Creates a new {@link WsnWsdl2Java} object.
    *
    * @param wsdlFiles DOCUMENT_ME
    * @param outputDir DOCUMENT_ME
    * @param classpath DOCUMENT_ME
    * @param options   DOCUMENT_ME
    *
    * @throws Exception DOCUMENT_ME
    */
   public WsnWsdl2Java( File[]           wsdlFiles,
                        File             outputDir,
                        File[]           classpath,
                        Wsdl2JavaOptions options )
   throws Exception
   {
      super( wsdlFiles, outputDir, classpath, options );
      initPortTypeInfoMap(  );
   }

   /**
    * Returns the name of the webapp for use in generating code.
    *
    * @return the webapp name.
    */
   protected String getWebbappName(  )
   {
      return WEBAPP_NAME;
   }

   /**
    * Adds to the {@link ServiceProperties} object by making a call to super and then adding anything impl specific to
    * be used by the VelocityContext for code generation.
    * <p/>
    * {@see Wsdl2Java#buildServiceProperties(ResourceDefinition resourceDefinition, File wsdlfile, File servicedir)}
    *
    * @param resourceDef the definition we are working on.
    * @param wsdlfile    the wsdl file.
    * @param servicedir  the directory we are generating to for this service.
    *
    * @return the updated ServiceProperties
    */
   protected ServiceProperties buildServiceProperties( ResourceDefinition resourceDef,
                                                       File               wsdlfile,
                                                       File               servicedir )
   {
      ServiceProperties serviceProps = super.buildServiceProperties( resourceDef, wsdlfile, servicedir );
      serviceProps.setNamespaceVersionHolder( WsnNamespaceVersionHolderImpl.class ); // make sure to add different ones as new versions come out

      // tell service props which properties we handle for the user...
      if ( isNotificationProducer( resourceDef ) )
      {
         serviceProps.addImplementedProperty( NotificationProducerPortType.PROP_QNAME_FIXED_TOPIC_SET );
         serviceProps.addImplementedProperty( NotificationProducerPortType.PROP_QNAME_TOPIC );
         serviceProps.addImplementedProperty( NotificationProducerPortType.PROP_QNAME_TOPIC_EXPRESSION_DIALECTS );
         serviceProps.addSpringBeanDefIncludeLocation( "templates/producer-spring-beans.xml.vm" );
      }

      return serviceProps;
   }

   /**
    * DOCUMENT_ME
    *
    * @param resourceDef DOCUMENT_ME
    * @param wsdlFile DOCUMENT_ME
    * @param serviceDir DOCUMENT_ME
    *
    * @throws Exception DOCUMENT_ME
    */
   protected void processTemplates( ResourceDefinition resourceDef,
                                    File               wsdlFile,
                                    File               serviceDir )
   throws Exception
   {
      super.processTemplates( resourceDef, wsdlFile, serviceDir );
      processPropValueChangeTopicSpaceTemplate( resourceDef, serviceDir );
   }

   /**
    * Provides a hook for adding directly to the Velocity Context.
    *
    * @param context
    *
    * @return VelocityContext
    */
   protected VelocityContext updateVelocityContext( VelocityContext    context,
                                                    ResourceDefinition resourceDef )
   {
      VelocityContext velocityContext = super.updateVelocityContext( context, resourceDef );
      if ( hasPropChange( resourceDef ) )
      {
         velocityContext.put( VCTY_CTX_PROP_CHANGE, "true" );
      }

      if ( hasResourceTermination( resourceDef ) )
      {
         velocityContext.put( VCTY_CTX_RESOURCE_TERMINATION, "true" );
      }

      if ( isNotificationProducer( resourceDef ) )
      {
         velocityContext.put( VCTY_CTX_NOTIF_PRODUCER, "true" );
      }

      velocityContext.put( VCTY_CTX_TOPICS_XML_RESOURCE_PATHS,
                           getPropValueChangeTopicsXmlResourcePaths( resourceDef ) );
      return velocityContext;
   }

   private boolean isNotificationProducer( ResourceDefinition resourceDef )
   {
      return resourceDef.implementsResourceCapability( org.apache.ws.notification.base.v2004_06.porttype.NotificationProducerPortType.NAME );
   }

   private List getPropValueChangeTopicsXmlResourcePaths( ResourceDefinition resourceDef )
   {
      Map  nsToPropNamesMap = buildNamespaceToPropertyLocalNamesMap( resourceDef );
      List resourcePaths = new ArrayList(  );
      for ( Iterator iterator = nsToPropNamesMap.keySet(  ).iterator(  ); iterator.hasNext(  ); )
      {
         String propNs          = (String) iterator.next(  );
         String javaPackageName = GenerationUtils.getJavaPackageName( propNs );
         String resourcePath    =
            javaPackageName.replace( '.', '/' ) + "/" + FILENAME_PROP_VALUE_CHANGE_TOPICS_XML;
         resourcePaths.add( resourcePath );
      }

      return resourcePaths;
   }

   private Map buildNamespaceToPropertyLocalNamesMap( ResourceDefinition resourceDef )
   {
      Map     nsToPropNamesMap = new HashMap(  );
      QName[] customPropNames = resourceDef.getCustomPropertyNames(  );
      for ( int i = 0; i < customPropNames.length; i++ )
      {
         QName propName  = customPropNames[i];
         Set   propNames;
         if ( nsToPropNamesMap.containsKey( propName.getNamespaceURI(  ) ) )
         {
            propNames = (Set) nsToPropNamesMap.get( propName.getNamespaceURI(  ) );
         }
         else
         {
            propNames = new HashSet(  );
            nsToPropNamesMap.put( propName.getNamespaceURI(  ),
                                  propNames );
         }

         propNames.add( propName.getLocalPart(  ) );
      }

      return nsToPropNamesMap;
   }

   private boolean hasPropChange( ResourceDefinition resourceDef )
   {
      if ( resourceDef.implementsResourceCapability( org.apache.ws.resource.properties.v2004_06.porttype.SetResourcePropertiesPortType.NAME )
           || resourceDef.implementsResourceCapability( org.apache.ws.resource.properties.v2004_11.porttype.DeleteResourcePropertiesPortType.NAME )
           || resourceDef.implementsResourceCapability( org.apache.ws.resource.properties.v2004_11.porttype.InsertResourcePropertiesPortType.NAME )
           || resourceDef.implementsResourceCapability( org.apache.ws.resource.properties.v2004_11.porttype.SetResourcePropertiesPortType.NAME )
           || resourceDef.implementsResourceCapability( org.apache.ws.resource.properties.v2004_11.porttype.UpdateResourcePropertiesPortType.NAME ) )
      {
         return true;
      }

      return false;
   }

   private boolean hasResourceTermination( ResourceDefinition resourceDef )
   {
      if ( resourceDef.implementsResourceCapability( org.apache.ws.resource.lifetime.v2004_06.porttype.ImmediateResourceTerminationPortType.NAME )
           || resourceDef.implementsResourceCapability( org.apache.ws.resource.lifetime.v2004_06.porttype.ScheduledResourceTerminationPortType.NAME )
           || resourceDef.implementsResourceCapability( org.apache.ws.resource.lifetime.v2004_11.porttype.ImmediateResourceTerminationPortType.NAME )
           || resourceDef.implementsResourceCapability( org.apache.ws.resource.lifetime.v2004_11.porttype.ScheduledResourceTerminationPortType.NAME ) )
      {
         return true;
      }

      return false;
   }

   private void initPortTypeInfoMap(  )
   {
      addPortType2JavaInfo( new NotificationConsumerPortType2JavaInfo(  ) );
      addPortType2JavaInfo( new NotificationProducerPortType2JavaInfo(  ) );
      addPortType2JavaInfo( new SubscriptionManagerPortType2JavaInfo(  ) );
   }

   private void processPropValueChangeTopicSpaceTemplate( ResourceDefinition resourceDef,
                                                          File               serviceDir )
   throws Exception
   {
      Map             nsToPropLocalNamesMap = buildNamespaceToPropertyLocalNamesMap( resourceDef );
      VelocityContext context = new VelocityContext(  );
      for ( Iterator iterator = nsToPropLocalNamesMap.keySet(  ).iterator(  ); iterator.hasNext(  ); )
      {
         String propNs = (String) iterator.next(  );
         context.put( "propertyNamespace", propNs );
         context.put( "propertyLocalNameSet",
                      nsToPropLocalNamesMap.get( propNs ) );
         File outputFile = new File( getPackageDir( propNs, serviceDir ),
                                     FILENAME_PROP_VALUE_CHANGE_TOPICS_XML );
         processTemplate( context, "templates/ResourcePropertyValueChangeTopics.xml.vm", outputFile );
      }
   }
}