/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.notification.topics.expression;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.notification.topics.Topic;
import org.apache.ws.notification.topics.TopicNamespaceRegistry;
import org.apache.ws.notification.topics.expression.impl.ConcreteTopicExpressionEvaluator;
import org.apache.ws.notification.topics.expression.impl.FullTopicExpressionEvaluator;
import org.apache.ws.notification.topics.expression.impl.SimpleTopicExpressionEvaluator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Resolves topic expressions over the set of topic namespaces stored in its topic
 * namespace registry.
 *
 * @see TopicNamespaceRegistry
 */
public class TopicExpressionEngine
{
   private static final Log       LOG               = LogFactory.getLog( TopicExpressionEngine.class.getName(  ) );
   private TopicNamespaceRegistry m_topicNsRegistry;
   private Map                    m_evaluators = new HashMap(  );

   /**
    * Creates a new {@link TopicExpressionEngine}.
    */
   public TopicExpressionEngine( TopicNamespaceRegistry topicNsRegistry )
   {
      this( topicNsRegistry, new HashMap(  ) );
   }

   /**
    * Creates a new {@link TopicExpressionEngine}.
    *
    * @param topicNsRegistry
    * @param topicExprEvaluatorMap
    */
   public TopicExpressionEngine( TopicNamespaceRegistry topicNsRegistry,
                                 Map                    topicExprEvaluatorMap )
   {
      m_topicNsRegistry    = topicNsRegistry;
      m_evaluators         = topicExprEvaluatorMap;
      if ( m_evaluators.isEmpty(  ) )
      {
         // defaults...
         registerEvaluator( new SimpleTopicExpressionEvaluator(  ) );
         registerEvaluator( new ConcreteTopicExpressionEvaluator(  ) );
         registerEvaluator( new FullTopicExpressionEvaluator(  ) );
      }
   }

   /**
    * DOCUMENT_ME
    *
    * @param dialect DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public synchronized TopicExpressionEvaluator getEvaluator( String dialect )
   {
      return (TopicExpressionEvaluator) m_evaluators.get( dialect );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public synchronized String[] getSupportedDialects(  )
   {
      return (String[]) m_evaluators.keySet(  ).toArray( new String[m_evaluators.size(  )] );
   }

   /**
    * DOCUMENT_ME
    *
    * @param topicExpr       DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    *
    * @throws org.apache.ws.notification.topics.expression.TopicPathDialectUnknownException
    *                                  DOCUMENT_ME
    * @throws TopicExpressionResolutionException
    *                                  DOCUMENT_ME
    * @throws InvalidTopicExpressionException
    *                                  DOCUMENT_ME
    * @throws TopicExpressionException DOCUMENT_ME
    * @throws TopicPathDialectUnknownException
    *                                  DOCUMENT_ME
    */
   public Topic[] evaluateTopicExpression( TopicExpression topicExpr )
   throws TopicPathDialectUnknownException, 
          TopicExpressionResolutionException, 
          InvalidTopicExpressionException, 
          TopicExpressionException
   {
      if ( topicExpr == null )
      {
         throw new IllegalArgumentException( "Specified TopicExpression was null." );
      }

      if ( topicExpr.getDialect(  ) == null )
      {
         throw new org.apache.ws.notification.topics.expression.TopicPathDialectUnknownException( "nullArgument"
                                                                                                  + "topicExpression.dialect" );
      }

      String                   dialect   = topicExpr.getDialect(  );
      TopicExpressionEvaluator evaluator = getEvaluator( dialect );

      if ( evaluator == null )
      {
         if ( LOG.isDebugEnabled(  ) )
         {
            LOG.debug( "Dialect not supported:" + dialect );
            LOG.debug( "Registered dialects are:" );
            Iterator keyIterator = m_evaluators.keySet(  ).iterator(  );
            Object   key = null;
            while ( keyIterator.hasNext(  ) )
            {
               key = keyIterator.next(  );
               LOG.debug( key );
            }

            LOG.debug( "key.equals(dialect): " + ( key.equals( dialect.toString(  ) ) ) );
            LOG.debug( "m_evaluators.containsKey(dialect): " + m_evaluators.containsKey( dialect ) );
            LOG.debug( "HashCode of key: " + String.valueOf( key.hashCode(  ) ) );
            LOG.debug( "HashCode of dialect: " + String.valueOf( dialect.toString(  ).hashCode(  ) ) );
            LOG.debug( "Object stored for key: " + m_evaluators.get( key ) );
         }

         throw new TopicPathDialectUnknownException( "The dialect '" + dialect + "' was not recognized." );
      }

      return evaluator.evaluate( m_topicNsRegistry, topicExpr );
   }

   /**
    * DOCUMENT_ME
    *
    * @param evaluator DOCUMENT_ME
    */
   public void registerEvaluator( TopicExpressionEvaluator evaluator )
   {
      LOG.debug( "Adding dialects for " + evaluator.getClass(  ).getName(  ) );
      String[] dialects = evaluator.getDialects(  );
      for ( int i = 0; i < dialects.length; i++ )
      {
         LOG.debug( "Adding dialect: " + dialects[i] );
         m_evaluators.put( dialects[i], evaluator );
      }
   }
}