/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.pubsub.emitter;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.pubsub.emitter.http.HttpEmitterTask;
import org.apache.ws.pubsub.i18n.Keys;
import org.apache.ws.pubsub.i18n.MessagesImpl;
import org.apache.ws.util.i18n.Messages;
import org.xml.sax.SAXException;
import javax.xml.soap.SOAPMessage;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.HashMap;

/**
 * This task is to be used to send notifications to client subscriber destinations.  This class contains
 * a factory method, {@link #createEmitterTask(SOAPMessage, URL)}, that creates a task specific to the
 * protocol for the given destination; this factory method must be used to create emitter tasks.
 * All subclasses to this class will be responsible for sending notifications over particular protocol
 * transports.  Each subclass must have a constructor that takes a SOAP message (the notification message itself)
 * and a destination URL. These constructors are called by this class' factory method.
 *
 * TODO (ips): refactor the code from this class into {@link SoapMessageEmitterEngine}
 */
public abstract class EmitterTask
   implements Runnable
{
   /** object used to obtain I18N messages */
   private static final Messages MSG = MessagesImpl.getInstance(  );

   /** object used to log messages */
   private static final Log LOG = LogFactory.getLog( EmitterTask.class );

   /**
    * Protocol emitter map: key is protocol name, value is corresponding {@link EmitterTask} class.
    */
   private static final HashMap PROTOCOL_EMITTER_MAP;

   static
   {
      PROTOCOL_EMITTER_MAP = new HashMap(  );
      PROTOCOL_EMITTER_MAP.put( HttpEmitterTask.PROTOCOL, HttpEmitterTask.class );

      //PROTOCOL_EMITTER_MAP.put( FileEmitterTask.PROTOCOL, FileEmitterTask.class );
      //PROTOCOL_EMITTER_MAP.put( HttpsNotificationEmitterTask.PROTOCOL, HttpsNotificationEmitterTask.class );
      //PROTOCOL_EMITTER_MAP.put( MailtoNotificationEmitterTask.PROTOCOL, MailtoNotificationEmitterTask.class );
   }

   /** The notification to send */
   private final SOAPMessage m_notification;

   /** The destination URL where the notification will be sent */
   private final URL m_destinationUrl;

   /**
    * Constructor to build the task.  This is protected to enforce that fact that only
    * {@link #createEmitterTask(javax.xml.soap.SOAPMessage, URL)}
    * should be used by clients to create these task objects.
    *
    * @param notif notification to send
    * @param url endpoint URL of notification consumer
    */
   protected EmitterTask( SOAPMessage notif,
                          URL         url )
   {
      m_notification      = notif;
      m_destinationUrl    = url;
   }

   /**
    * Configures a notification emitter to handle emitting notifications
    * to consumer EPRs with addresses of the specified protocol type.
    *
    * @param protocol a URL protocol (e.g. http, https, or mailto)
    * @param emitterClass the emitter class - must extend {@link EmitterTask}
    */
   public static void addEmitter( String protocol,
                                  Class  emitterClass )
   {
      PROTOCOL_EMITTER_MAP.put( protocol, emitterClass );
   }

   /**
    * Factory method that creates the appropriate emitter task based on the
    * protocol required to send the notification.
    *
    * @param notif notification to send
    * @param destination notification is to be sent to this destination endpoint
    *
    * @return the task to be used to send the notification
    *
    * @throws EmitterException if failed to create the task
    */
   public static EmitterTask createEmitterTask( SOAPMessage notif,
                                                URL         destination )
   throws EmitterException
   {
      EmitterTask task;
      Class       task_class = null;

      LOG.debug( MSG.getMessage( Keys.CREATING_NOTIF_EMITTER, destination ) );

      try
      {
         task_class = (Class) PROTOCOL_EMITTER_MAP.get( destination.getProtocol(  ) );

         if ( task_class != null )
         {
            Class[] sig = new Class[]
                          {
                             SOAPMessage.class,
                             URL.class
                          };
            Object[] params = new Object[]
                              {
                                 notif,
                                 destination
                              };

            task = (EmitterTask) task_class.getConstructor( sig ).newInstance( params );
         }
         else
         {
            String err = MSG.getMessage( Keys.UNSUPPORTED_PROTOCOL, destination, notif );
            LOG.error( err );
            throw new EmitterException( err );
         }
      }
      catch ( InstantiationException ie )
      {
         LOG.error( MSG.getMessage( Keys.BAD_TASK_CLASS_IS_ABSTRACT,
                                    task_class.getName(  ),
                                    ie ) );
         throw new EmitterException( ie );
      }
      catch ( IllegalAccessException iae )
      {
         LOG.error( MSG.getMessage( Keys.BAD_TASK_CLASS_ACCESS_DENIED,
                                    task_class.getName(  ),
                                    iae ) );
         throw new EmitterException( iae );
      }
      catch ( InvocationTargetException ite )
      {
         LOG.error( MSG.getMessage( Keys.TASK_CON_EXCEPTION,
                                    task_class.getName(  ),
                                    ite ) );
         throw new EmitterException( ite );
      }
      catch ( NoSuchMethodException nsme )
      {
         LOG.error( MSG.getMessage( Keys.BAD_TASK_CLASS_NO_CON,
                                    task_class.getName(  ),
                                    nsme ) );
         throw new EmitterException( nsme );
      }

      return task;
   }

   /**
    * This template method runs the task's thread and will call the appropriate
    * emitter task's {@link #emit()} method to send the notification.
    *
    * @see java.lang.Runnable#run()
    */
   public void run(  )
   {
      LOG.debug( MSG.getMessage( Keys.EMITTING_NOTIFICATION, m_destinationUrl, m_notification ) );
      try
      {
         emit(  );
      }
      catch ( Throwable t )
      {
         if ( !( t.getCause(  ) instanceof SAXException ) )
         {
            LOG.error( MSG.getMessage( Keys.EMISSION_FAILURE,
                                       this.getClass(  ),
                                       m_notification,
                                       t.getMessage(  ) ) );
         }
         else
         {
            // Axis throws a SAXException if the notification consumer returns a response with an invalid HTTP body,
            // which we can basically ignore.
            LOG.trace( MSG.getMessage( Keys.EMISSION_FAILURE,
                                       this.getClass(  ),
                                       m_notification,
                                       t.getCause(  ) ) );
         }
      }
   }

   /**
    * Returns the destination URL where the notification will be sent
    *
    * @return the destination URL where the notification will be sent
    */
   protected URL getDestinationUrl(  )
   {
      return m_destinationUrl;
   }

   /**
    * Returns the actual notification object that is to be emitted.
    *
    * @return the actual notification object that is to be emitted.
    */
   protected SOAPMessage getNotification(  )
   {
      return m_notification;
   }

   /**
    * Method that will actually send the notification to the destination via
    * the appropriate protocol transport.
    *
    * @throws EmitterException on error attempting to send notification
    */
   protected abstract void emit(  )
   throws EmitterException;
}